# -*- coding: utf-8 -*-
from enum import Enum


class TraceEvent(Enum):
    line = 'line'
    call = 'call'
    return_ = 'return'
    exception = 'exception'

    # these are included for completeness but will probably not be used
    c_call = 'c_call'
    c_return = 'c_return'
    c_exception = 'c_exception'

    def __str__(self):
        # maxlen = max(len(v.value) for v in self.__class__.__dict__.values() if isinstance(v, self.__class__))
        # return ' ' * (maxlen - len(self.value)) + self.value
        return self.value



# -*- coding: utf-8 -*-
import ast
import logging
from typing import TYPE_CHECKING

from .attr_symbols import get_attribute_symbol_chain
from .mixins import SkipUnboundArgsMixin, VisitListsMixin

if TYPE_CHECKING:
    from .attr_symbols import AttributeSymbolChain
    from typing import KeysView, List, Set, Union

logger = logging.getLogger(__name__)


# TODO: have the logger warnings additionally raise exceptions for tests
class PreCheck(ast.NodeVisitor):

    def __init__(self):
        self.safe_set: Set[Union[str, AttributeSymbolChain]] = set()

    def __call__(self, module_node: ast.Module, name_set: 'KeysView[str]'):
        """
        This function should be called when we want to precheck an ast.Module. For
        each line/block of the cell we first run the check of new assignments, then
        we obtain all the names. In these names, we put the ones that are user
        defined and not in the safe_set into the return check_set for further
        checks.
        """
        check_set = set()
        for node in module_node.body:
            self.visit(node)
            for name in get_all_names(node):
                if name not in self.safe_set:
                    check_set.add(name)
        return check_set

    # In case of assignment, we put the new assigned variable into a safe_set
    # to indicate that we know for sure it won't have stale dependency.  Note
    # that node.targets might contain multiple ast.Name node in the case of
    # "a = b = 3", so we go through each node in the targets.  Also note that
    # `target` would be an ast.Tuple node in the case of "a,b = 3,4". Thus
    # we need to break the tuple in that case.
    def visit_Assign(self, node: ast.Assign):
        for target_node in node.targets:
            if isinstance(target_node, ast.Tuple):
                for element_node in target_node.elts:
                    if isinstance(element_node, ast.Name):
                        self.safe_set.add(element_node.id)
            else:
                self.visit_Assign_or_AugAssign_target(target_node)

    def visit_AugAssign(self, node: ast.AugAssign):
        self.visit_Assign_or_AugAssign_target(node.target)

    def visit_Assign_or_AugAssign_target(self, target_node: 'Union[ast.Attribute, ast.Name, ast.Subscript, ast.expr]'):
        ignore_node_types = (ast.Subscript,)
        if isinstance(target_node, ignore_node_types):
            return
        elif isinstance(target_node, ast.Name):
            self.safe_set.add(target_node.id)
        elif isinstance(target_node, ast.Attribute):
            self.safe_set.add(get_attribute_symbol_chain(target_node))
        else:
            logger.warning('unsupported type for node %s' % target_node)

    # We also put the name of new functions in the safe_set
    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.safe_set.add(node.name)

    def visit_For(self, node: ast.For):
        # Case "for a,b in something: "
        if isinstance(node.target, ast.Tuple):
            for name_node in node.target.elts:
                if isinstance(name_node, ast.Name):
                    self.safe_set.add(name_node.id)
                else:
                    logger.warning('unsupported type for node %s' % name_node)
        # case "for a in something"
        elif isinstance(node.target, ast.Name):
            self.safe_set.add(node.target.id)
        else:
            logger.warning('unsupported type for node %s' % node.target)

        # Then we keep doing the visit for the body of the loop.
        for line in node.body:
            self.visit(line)


def precheck(code: 'Union[ast.Module, str]', name_set: 'KeysView[str]'):
    if isinstance(code, str):
        code = ast.parse(code)
    return PreCheck()(code, name_set)


# Call GetAllNames()(ast_tree) to get a set of all names appeared in ast_tree.
# Helper Class
class GetAllNames(SkipUnboundArgsMixin, VisitListsMixin, ast.NodeVisitor):
    def __init__(self):
        self.name_set: Set[Union[str, AttributeSymbolChain]] = set()

    def __call__(self, node: ast.AST):
        self.visit(node)
        return self.name_set

    def visit_Name(self, node: ast.Name):
        self.name_set.add(node.id)

    # We overwrite FunctionDef because we don't need to check names in the body of the definition.
    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.visit(node.args)

    def visit_ClassDef(self, node: ast.ClassDef):
        self.generic_visit(node.bases)
        self.generic_visit(node.decorator_list)

    def visit_Attribute(self, node: ast.Attribute):
        self.name_set.add(get_attribute_symbol_chain(node))


def get_all_names(node: ast.AST):
    return GetAllNames()(node)



# -*- coding: utf-8 -*-
import logging

from .utils import make_safety_fixture, skipif_known_failing


logging.basicConfig(level=logging.ERROR)


# Reset dependency graph before each test
_safety_fixture, _safety_state, run_cell = make_safety_fixture()


def stale_detected():
    return _safety_state[0].test_and_clear_detected_flag()


def assert_bool(val, msg=''):
    assert val, str(msg)


def assert_detected(msg=''):
    assert_bool(stale_detected(), msg=msg)


def assert_false_positive(msg=''):
    """
    Same as `assert_detected` but asserts a false positive.
    Helps with searchability of false positives in case we want to fix these later.
    """
    return assert_detected(msg=msg)


def assert_not_detected(msg=''):
    assert_bool(not stale_detected(), msg=msg)


def assert_false_negative(msg=''):
    """
    Same as `assert_not_detected` but asserts a false negative.
    Helps with searchability of false negatives in case we want to fix these later.
    """
    return assert_not_detected(msg=msg)


def test_simplest():
    run_cell('a = 1')
    run_cell('b = a + 1')
    run_cell('a = 3')
    run_cell('logging.info(b)')
    assert_detected('should have detected b has stale dep on old a')


def test_subscript_dependency():
    run_cell('lst = [0, 1, 2]')
    run_cell('x = 5')
    run_cell('y = x + lst[0]')
    run_cell('lst[0] = 10')
    run_cell('logging.info(y)')
    assert_false_positive('y depends on stale lst[0]')


def test_long_chain():
    run_cell('a = 1')
    run_cell('b = a + 1')
    run_cell('c = b + 1')
    run_cell('d = c + 1')
    run_cell('e = d + 1')
    run_cell('f = e + 1')
    assert_not_detected('everything OK so far')
    run_cell('a = 2')
    run_cell('logging.info(f)')
    assert_detected('f has stale dependency on old value of a')


def test_redef_after_stale_use():
    run_cell('a = 1')
    run_cell('b = a + 1')
    run_cell('a = 3')
    run_cell("""
logging.info(b)
b = 7
""")
    assert_detected('b has stale dependency on old value of a')


def test_subscript_dependency_fp():
    run_cell('lst = [0, 1, 2]')
    run_cell('x = 5')
    run_cell('y = x + lst[0]')
    run_cell('lst[1] = 10')
    run_cell('logging.info(y)')
    assert_false_positive('false positive on unchanged lst[0] but OK since fine-grained detection hard')


# simple test about the basic assignment
def test_basic_assignment():
    run_cell('a = 1')
    run_cell('b = 2')
    run_cell('c = a+b')
    run_cell('d = c+1')
    run_cell('logging.info(a,b,c,d)')
    # redefine a here but not c and d
    run_cell('a = 7')
    run_cell('logging.info(a,b,c,d)')
    assert_detected("Did not detect that c's reference was changed")

    run_cell('c = a+b')
    run_cell('logging.info(a,b,c,d)')
    assert_detected("Did not detect that d's reference was changed")

    run_cell('d = c+1')
    run_cell('logging.info(a,b,c,d)')
    assert_not_detected("There should be no more dependency issue")


# redefined function example from the project prompt
def test_redefined_function_in_list():
    run_cell("""
def foo():
    return 5

def bar():
    return 7
""")
    run_cell("""
funcs_to_run = [foo,bar]
""")
    run_cell("""
accum = 0
for f in funcs_to_run:
    accum += f()
logging.info(accum)
""")
    
    # redefine foo here but not funcs_to_run
    run_cell("""
def foo():
    return 10

def bar():
    return 7
""")
    run_cell("""
accum = 0
for f in funcs_to_run:
    accum += f()
logging.info(accum)
""")
    assert_detected("Did not detect that funcs_to_run's reference was changed")

    run_cell("""
funcs_to_run = [foo,bar]
""")
    run_cell("""
accum = 0
for f in funcs_to_run:
    accum += f()
logging.info(accum)
""")
    assert_not_detected("There should be no more dependency issue")


# like before but the function is called in a list comprehension
def test_redefined_function_for_funcall_in_list_comp():
    run_cell("""
def foo():
    return 5

def bar():
    return 7
""")
    run_cell('retvals = [foo(), bar()]')
    run_cell("""
accum = 0
for ret in retvals:
    accum += ret
logging.info(accum)
""")

    # redefine foo here but not funcs_to_run
    run_cell("""
def foo():
    return 10

def bar():
    return 7
""")
    run_cell('logging.info(accum)')
    assert_detected('Did not detect stale dependency of `accum` on `foo` and `bar`')


# like before but we run the list through a function before iterating
def test_redefined_function_for_funcall_in_modified_list_comp():
    run_cell("""
def foo():
    return 5

def bar():
    return 7
""")
    run_cell('retvals = tuple([foo(), bar()])')
    run_cell("""
accum = 0
# for ret in map(lambda x: x * 5, retvals):
for ret in retvals:
    accum += ret
logging.info(accum)
""")

    # redefine foo here but not funcs_to_run
    run_cell("""
def foo():
    return 10

def bar():
    return 7
""")
    run_cell('logging.info(accum)')
    assert_detected('Did not detect stale dependency of `accum` on `foo` and `bar`')


def test_for_loop_with_map():
    run_cell("""
accum = 0
foo = [1, 2, 3, 4, 5]
for ret in map(lambda x: x * 5, foo):
    accum += ret
logging.info(accum)
""")
    assert_not_detected('no stale dep foo -> accum')
    run_cell('foo = [0]')
    run_cell('logging.info(accum)')
    assert_detected('did not detect stale dep foo -> accum')


def test_redefined_function_over_list_comp():
    run_cell("""
def foo():
    return 5

def bar():
    return 7

def baz(lst):
    return map(lambda x: 3*x, lst)
""")
    run_cell('retvals = baz([foo(), bar()])')
    run_cell("""
accum = 0
for ret in map(lambda x: x * 5, retvals):
    accum += ret
""")
    run_cell("""
def baz(lst):
    return map(lambda x: 7*x, lst)
""")
    run_cell('logging.info(accum)')
    assert_detected('Did not detect stale dependency of `accum` on `baz`')


# like before but the function is called in a tuple comprehension
def test_redefined_function_for_funcall_in_tuple_comp():
    run_cell("""
def foo():
    return 5

def bar():
    return 7
""")
    run_cell('retvals = (foo(), bar())')
    run_cell("""
accum = 0
for ret in retvals:
    accum += ret
logging.info(accum)
""")

    # redefine foo here but not funcs_to_run
    run_cell("""
def foo():
    return 10

def bar():
    return 7
""")
    run_cell('logging.info(accum)')
    assert_detected('Did not detect stale dependency of `accum` on `foo` and `bar`')


# Tests about variables that have same name but in different scope.
# There shouldn't be any extra dependency because of the name.
def test_variable_scope():
    run_cell("""
def func():
    x = 6
""")
    run_cell('x = 7')
    run_cell('y = x')
    run_cell('z = func')
    run_cell('logging.info(y,z())')

    # change x inside of the function, but not x outside of the function
    run_cell('def func():\n    x = 10')
    run_cell('logging.info(y,z())')
    assert_detected("Did not detect the dependency change in the function")

    run_cell('y = x')
    run_cell('logging.info(y,z())')
    assert_detected("Updating y should not solve the dependency change inside of function func")

    run_cell('z = func')
    run_cell('logging.info(y,z())')
    assert_not_detected("Updating z should solve the problem")


def test_variable_scope_2():
    run_cell('def func():\n    x = 6')
    run_cell('x = 7')
    run_cell('y = x')
    run_cell('z = func')
    run_cell('logging.info(y,z())')

    # change x outside of the function, but not inside of the function
    run_cell('x = 10')
    run_cell('logging.info(y,z())')
    assert_detected("Did not detect the dependency change outside of the function")

    run_cell('z = func')
    run_cell('logging.info(y,z())')
    assert_detected("Updating z should not solve the dependency change outside of function")

    run_cell('y = x')
    run_cell('logging.info(y,z())')
    assert_not_detected("Updating y should solve the problem")


def test_default_args():
    run_cell("""
x = 7
def foo(y=x):
    return y + 5
""")
    run_cell('a = foo()')
    assert_not_detected()
    run_cell('x = 10')
    assert_not_detected()
    run_cell('b = foo()')
    assert_detected("Should have detected stale dependency of fn foo() on x")


def test_same_pointer():
    # a and b are actually pointing to the same thing
    run_cell('a = [7]')
    run_cell('b = a')
    run_cell('c = b + [5]')

    run_cell('a[0] = 8')
    run_cell('logging.info(b)')
    assert_not_detected('`b` is an alias of `a`, updating a should automatically update `b` as well')
    run_cell('logging.info(c)')
    assert_detected('`c` does not point to the same thing as `a` or `b`, thus there is a stale dependency here')


def test_func_assign_objs():
    run_cell("""
a = [1]
b = [1]
c = [2]
d = [3]
def func(x, y=a):
    e = [c[0] + d[0]]
    f = [x[0] + y[0]]
    return f
""")
    run_cell('z = func(c)')
    run_cell('a = [4]')
    run_cell('logging.info(z[0])')
    assert_detected("Should have detected stale dependency of fn func on a")
    run_cell("""
def func(x, y=a):
    logging.info(b[0])
    e = [c[0] + d[0]]
    f = [x[0] + y[0]]
    return f
z = func(c)
""")
    run_cell('logging.info(z[0])')
    assert_not_detected()
    run_cell('c = [3]')
    run_cell('logging.info(z[0])')
    assert_detected("Should have detected stale dependency of z on c")
    run_cell('z = func(c)')
    run_cell('logging.info(z[0])')
    assert_not_detected()
    run_cell('b = [4]')
    run_cell('d = [1]')
    assert_not_detected("Changing b and d should not affect z")


def test_func_assign_ints():
    run_cell("""
a = 1
b = 1
c = 2
d = 3
def func(x, y=a):
    e = c + d
    f = x + y
    return f
""")
    run_cell('z = func(c)')
    run_cell('a = 4')
    run_cell('logging.info(z)')
    assert_detected("Should have detected stale dependency of fn func on a")
    run_cell("""
def func(x, y=a):
    logging.info(b)
    e = c + d
    f = x + y
    return f
z = func(c)
""")
    run_cell('logging.info(z)')
    assert_not_detected()
    run_cell('c = 3')
    run_cell('logging.info(z)')
    assert_detected("Should have detected stale dependency of z on c")
    run_cell('z = func(c)')
    run_cell('logging.info(z)')
    assert_not_detected()
    run_cell('b = 4')
    run_cell('d = 1')
    assert_not_detected("Changing b and d should not affect z")


def test_func_assign_helper_func():
    run_cell("""
x = 3
a = 4
def f():
    def g():
        logging.info(a)
        return x
    return g()
y = f()
""")
    run_cell('x = 4')
    run_cell('logging.info(y)')
    assert_detected("Should have detected stale dependency of y on x")
    run_cell('y = f()')
    run_cell('logging.info(y)')
    assert_not_detected()
    run_cell('a = 1')
    run_cell('logging.info(y)')
    assert_not_detected("Changing a should not affect y")


def test_func_assign_helper_func_2():
    run_cell("""
x = 3
a = 4
def f():
    def g():
        logging.info(a)
        return x
    return g
y = f()()
""")
    run_cell('x = 4')
    run_cell('logging.info(y)')
    assert_detected("Should have detected stale dependency of y on x")


def test_branching():
    run_cell('y = 7')
    run_cell('x = y + 3')
    run_cell("""
if True:
    b = 5
else:
    y = 7
""")
    run_cell('logging.info(x)')
    assert_not_detected('false positive on unchanged y')


def test_branching_2():
    run_cell('y = 7')
    run_cell('x = y + 3')
    run_cell("""
if False:
    b = 5
else:
    y = 7
""")
    run_cell('logging.info(x)')
    assert_detected('x depends on stale y')


def test_attributes():
    run_cell("""
class Foo(object):
    def __init__(self, x):
        self.x = x
""")
    run_cell('x = Foo(5)')
    run_cell('y = x.x + 5')
    run_cell('x.x = 8')
    run_cell('logging.info(y)')
    assert_detected('y depends on stale attrval x.x')


def test_attributes_2():
    run_cell("""
class Foo(object):
    def __init__(self, x):
        self.x = x
""")
    run_cell('x = Foo(5)')
    run_cell('y = x.x + 5')
    run_cell('x = 8')
    run_cell('logging.info(y)')
    assert_detected('y depends on stale x')


def test_attributes_3():
    run_cell("""
class Foo(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y
""")
    run_cell('foo = Foo(5, 6)')
    run_cell('bar = Foo(7, 8)')
    run_cell('y = bar.x + 5')
    run_cell('foo.x = 8')
    run_cell('logging.info(y)')
    assert_not_detected('y does not depend on updated attrval foo.y')


def test_stale_use_of_attribute():
    run_cell("""
class Foo(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y
""")
    run_cell('foo = Foo(5, 6)')
    run_cell('bar = Foo(7, 8)')
    run_cell('foo.x = bar.x + bar.y')
    run_cell('bar.y = 42')
    run_cell('logging.info(foo.x)')
    assert_detected('`foo.x` depends on stale `bar.y`')


def test_attr_manager_active_scope_resets():
    run_cell("""
y = 10
class Foo(object):
    def f(self):
        y = 10
        return y
def f():
    return y
""")
    run_cell('foo = Foo()')
    # if the active scope doesn't reset after done with foo.f(),
    # it will think the `y` referred to by f() is the one in Foo.f's scope.
    run_cell('x = foo.f() + f()')
    run_cell('y = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale `y`')


def test_attr_manager_active_scope_with_property():
    run_cell("""
y = 10
class Foo(object):
    @property
    def f(self):
        y = 11
        return y
""")
    run_cell('foo = Foo()')
    # if the active scope doesn't reset after done with foo.f(),
    # it will think the `y` referred to by f() is the one in Foo.f's scope.
    run_cell('x = foo.f')
    run_cell('y = 42')
    run_cell('logging.info(x)')
    assert_not_detected('`x` independent of outer `y`')


def test_namespace_scope_resolution():
    run_cell("""
y = 42
class Foo(object):
    y = 10
    @property
    def foo(self):
        return y
""")
    run_cell('foo = Foo()')
    run_cell('x = foo.foo')
    run_cell('Foo.y = 99')
    run_cell('logging.info(x)')
    assert_not_detected('`x` should not have dependency on `Foo.y`')


def test_long_chain_attribute():
    run_cell("""
class Foo(object):
    shared = 99
    def __init__(self, x, y):
        self.x = x
        self.y = y + self.shared
        
    class Bar(object):
        def __init__(self, a, b):
            self.a = a
            self.b = b
            
        def foo(self):
            return Foo(self.a, self.b)
            
        def bar(self):
            return Foo
""")
    run_cell('foo = Foo(5, 6)')
    run_cell('bar = Foo.Bar(7, 8)')
    run_cell('foo.x = 42')
    run_cell('logging.info(bar.a)')
    assert_not_detected()
    run_cell('Foo.Bar(9, 10).foo().shared = 100')
    run_cell('logging.info(foo.y)')
    assert_not_detected('we mutated `shared` on obj and not on `Foo` (the one on which `foo.y` depends)')
    run_cell('Foo.Bar(9, 10).bar().shared = 100')
    run_cell('logging.info(foo.y)')
    assert_detected('we mutated a shared value on which `foo.y` depends')


def test_numpy_subscripting():
    run_cell('import numpy as np')
    run_cell('x = np.zeros(5)')
    run_cell('y = x[3] + 5')
    run_cell('x[3] = 2')
    run_cell('logging.info(y)')
    assert_detected('y depends on stale x[3]')


def test_subscript_sensitivity():
    run_cell('lst = list(range(5))')
    run_cell('i = 0')
    run_cell('lst[i] = 10')
    run_cell('i = 1')
    run_cell('logging.info(lst)')
    assert_detected('lst depends on stale i')


def test_subscript_adds():
    run_cell('x = 42')
    run_cell("""
d = {
    'foo': x,
    'bar': 77
}
""")
    run_cell('d["bar"] = 99')
    run_cell('x = 100')
    run_cell('logging.info(d)')
    assert_detected('`d` depends on stale `x`')


def test_list_mutation():
    run_cell('lst = list(range(5))')
    run_cell('x = 42')
    run_cell('asdf = []')
    run_cell('asdf.append(lst.append(x))')
    run_cell('x = 43')
    run_cell('logging.info(lst)')
    assert_detected('lst depends on stale x')


def test_list_mutation_2():
    run_cell('lst = list(range(5))')
    run_cell('x = lst + [42, 43]')
    run_cell('lst.append(99)')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale `lst`')


def test_lazy_class_scope_resolution():
    run_cell("""
class Foo(object):
    shared = 99
    def __init__(self, x):
        self.x = x
""")
    run_cell('foo = Foo(10)')
    run_cell('y = 11')
    run_cell('Foo.shared = y + 42')
    run_cell('y = 12')
    run_cell('logging.info(foo.shared)')
    assert_detected('`foo.shared` should point to same DataCell as `Foo.shared` and thus also has stale dep')
    run_cell('foo.shared = 89')
    run_cell('logging.info(Foo.shared)')
    assert_detected('Even though we refreshed `foo.shared`, this '
                    'has no bearing on the original class member `Foo.shared`')


def test_numpy_subscripting_fp():
    run_cell('import numpy as np')
    run_cell('x = np.zeros(5)')
    run_cell('y = x[3] + 5')
    run_cell('x[0] = 2')
    run_cell('logging.info(y)')
    assert_false_positive('false positive on changed x[3] but OK since fine-grained detection hard')


def test_old_format_string():
    run_cell('a = 5\nb = 7')
    run_cell('expr_str = "{} + {} = {}".format(a, b, a + b)')
    run_cell('a = 9')
    run_cell('logging.info(expr_str)')
    assert_detected('`expr_str` depends on stale `a`')


def test_old_format_string_kwargs():
    run_cell('a = 5\nb = 7')
    run_cell('expr_str = "{a} + {b} = {total}".format(a=a, b=b, total=a + b)')
    run_cell('a = 9')
    run_cell('logging.info(expr_str)')
    assert_detected('`expr_str` depends on stale `a`')


def test_new_format_string():
    run_cell('a = 5\nb = 7')
    run_cell('expr_str = f"{a} + {b} = {a+b}"')
    run_cell('a = 9')
    run_cell('logging.info(expr_str)')
    assert_detected('`expr_str` depends on stale `a`')


def test_scope_resolution():
    run_cell("""
def f(x):
    def g(x):
        return 2 * x
    return g(x) + 8
""")
    run_cell('x = 7')
    run_cell('y = f(x)')
    run_cell('x = 8')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `x`')


def test_scope_resolution_2():
    run_cell("""
def g(x):
    return 2 * x
def f(x):
    return g(x) + 8
""")
    run_cell('x = 7')
    run_cell('y = f(x)')
    run_cell('x = 8')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `x`')


def test_funcall_kwarg():
    run_cell("""
def f(y):
    return 2 * y + 8
""")
    run_cell('x = 7')
    run_cell('z = f(y=x)')
    run_cell('x = 8')
    run_cell('logging.info(z)')
    assert_detected('`z` depends on stale `x`')


def test_funcall_kwarg_2():
    run_cell("""
def f(y):
    return 2 * y + 8
""")
    run_cell('x = 7')
    run_cell('y = f(y=x)')
    run_cell('x = 8')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `x`')


def test_funcall_kwarg_3():
    run_cell("""
def f(x):
    return 2 * x + 8
""")
    run_cell('x = 7')
    run_cell('y = f(x=x)')
    run_cell('x = 8')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on stale `x`')


def test_funcall_kwarg_4():
    run_cell("""
def f(x):
    return 2 * x + 8
""")
    run_cell('x = 7')
    run_cell('x = f(x=x)')
    run_cell('x = 8')
    run_cell('logging.info(x)')
    assert_not_detected('`x` is overriden so should not be stale')


def test_single_line_dictionary_literal():
    run_cell('foo = 5')
    run_cell('bar = 6')
    run_cell('d = {foo: bar, "pi": 42,}')
    run_cell('bar = 7')
    run_cell('logging.info(d)')
    assert_detected('`d` depends on stale `bar`')


@skipif_known_failing
def test_single_line_dictionary_literal_fix_stale_deps():
    run_cell('foo = 5')
    run_cell('bar = 6')
    run_cell('d = {foo: bar, "pi": 42,}')
    run_cell('bar = 7')
    run_cell('logging.info(d)')
    assert_detected('`d` depends on stale `bar`')
    run_cell('d[foo] = bar')
    assert_not_detected('`d`s stale dep fixed')
    run_cell('foo = 8')
    assert_detected('`d` depends on stale `foo`')
    run_cell('d[foo] = bar')
    assert_not_detected('`d`s stale dep fixed')


def test_multiline_dictionary_literal():
    run_cell('foo = 5')
    run_cell('bar = 6')
    run_cell("""
d = {
    foo: bar,
    'pi': 42,
}
""")
    run_cell('bar = 7')
    run_cell('logging.info(d)')
    assert_detected('`d` depends on stale `bar`')


def test_exception():
    run_cell('lst = list(range(5))')
    run_cell('x = 6')
    run_cell("""
try:
    lst[x] = 42
except:
    lst[0] = 42
""")
    run_cell('x = 7')
    run_cell('logging.info(lst)')
    assert_not_detected('lst should be independent of x due to exception')


def test_for_loop_binding():
    run_cell('a = 0')
    run_cell('b = 1')
    run_cell('c = 2')
    run_cell('lst = [a, b, c]')
    run_cell("""
for i in lst:
    pass
""")
    run_cell('a = 3')
    run_cell('logging.info(i)')
    assert_false_positive('`i` should not depend on `a` at end of for loop but this is hard')


@skipif_known_failing
def test_for_loop_literal_binding():
    run_cell('a = 0')
    run_cell('b = 1')
    run_cell('c = 2')
    run_cell("""
for i in [a, b, c]:
    pass
""")
    run_cell('a = 3')
    run_cell('logging.info(i)')
    assert_not_detected('`i` should not depend on `a` at end of for loop')


def test_same_cell_redefine():
    run_cell('a = 0')
    run_cell("""
b = a + 1
a = 42
""")
    run_cell('logging.info(b)')
    assert_not_detected('`b` should not be considered as having stale dependency since `a` changed in same cell as `b`')


@skipif_known_failing
def test_multiple_stmts_in_one_line():
    run_cell('a = 1; b = 2')
    run_cell('x = a + b')
    run_cell('a = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale value of `a`')


def test_multiple_stmts_in_one_line_2():
    run_cell('a = 1; b = 2')
    run_cell('x = a + b')
    run_cell('b = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale value of `a`')


def test_line_magic():
    run_cell("""
%lsmagic
%lsmagic
%lsmagic
%lsmagic
%lsmagic
a = 0
""")
    run_cell("""
%lsmagic
%lsmagic
%lsmagic
%lsmagic
x = a + 1
""")
    run_cell("""
%lsmagic
%lsmagic
%lsmagic
a = 42
""")
    run_cell("""
%lsmagic
%lsmagic
logging.info(x)
""")
    assert_detected('`x` depends on stale value of `a`')
    run_cell("""
%lsmagic
logging.info(x)
%lsmagic
""")
    assert_detected('`x` depends on stale value of `a`')


def test_exception_stack_unwind():
    import builtins
    safety_state = '_safety_state'
    test_passed = 'test_passed'
    setattr(builtins, safety_state, _safety_state[0])
    setattr(builtins, test_passed, True)
    test_passed = "'" + test_passed + "'"

    def assert_stack_size(size):
        return ';'.join([
            f'can_pass = len({safety_state}.trace_state.stack) == {size} '
            f'and len({safety_state}.attr_trace_manager.stack) == {size}',
            f'setattr(builtins, {test_passed}, getattr(builtins, {test_passed}) and can_pass)'
        ])
    try:
        run_cell(f"""
import builtins
import numpy as np
{assert_stack_size(0)}
def f():
    {assert_stack_size(1)}
    def g():
        {assert_stack_size(2)}
        def h():
            {assert_stack_size(3)}
            return np.loadtxt('does-not-exist.txt')
        return h()
    try:
        return g()
    except:
        {assert_stack_size(1)}
f()
{assert_stack_size(0)}
""")
    finally:
        delattr(builtins, safety_state)
    test_passed = test_passed.strip("'")
    try:
        assert getattr(builtins, test_passed), 'unexpected stack size somewhere'
    finally:
        delattr(builtins, test_passed)


def test_throwing_statements_do_not_track_deps():
    run_cell("""
z = 10
def foo():
    def bar():
        raise ValueError('foo!')
    return bar() + z
x = 0
y = x + 1
""")
    run_cell("""
try:
    x = 42 + foo()
except:
    pass
""")
    run_cell('logging.info(y)')
    assert_not_detected('no stale dep for `y` because update on `x` threw exception')
    run_cell('z = 99')
    run_cell('logging.info(x)')
    assert_not_detected('no stale dep for `x` because it is indep of `z` (attempted dep add threw)')


def test_attr_dep_from_somewhere_else():
    run_cell('import sys')
    run_cell('sys.path.append("./test")')
    run_cell('import fake')
    run_cell('fake.y = 7')
    run_cell('x = fake.y + 1')
    run_cell('fake.y = 42')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on old value of `fake.y`')


def test_attr_use_from_somewhere_else():
    run_cell('import sys')
    run_cell('sys.path.append("./test")')
    run_cell('import fake')
    run_cell('x = 7')
    run_cell('fake.y = x + 1')
    run_cell('x = 42')
    run_cell('logging.info(fake.y)')
    assert_detected('`fake.y` depends on old value of `x`')


def test_class_assignment():
    run_cell("""
class Foo:
    def __init__(self):
        self.y = 99

Bar = Foo
foo = Bar()
x = 7
""")
    run_cell('foo.y = x + 1')
    run_cell('x = 42')
    run_cell('logging.info(foo.y)')
    assert_detected('`foo.y` depends on stale `x`')
    run_cell('foo.y = 10')
    run_cell('x = foo.y + 1')
    run_cell('foo.y = 12')
    run_cell('logging.info(x)')
    assert_detected('`x` depends on stale `foo.y`')


def test_no_class_false_positives():
    run_cell('x = 7')
    run_cell('y = x + 1')
    run_cell('x = 42')
    run_cell("""
try:
    class Foo:
        print(y)
except:
    pass
""")
    assert_not_detected('x inside class scope is different')


def test_tuple_unpack_simple():
    run_cell('x, y = 0, 1')
    run_cell('a, b = x + 2, y + 3')
    run_cell('x, y = 42, 43')
    run_cell('logging.info(a)')
    assert_detected('`a` depends on stale `x`')
    run_cell('logging.info(b)')
    assert_detected('`b` depends on stale `y`')


@skipif_known_failing
def test_tuple_unpack_hard():
    run_cell('x, y = 0, 1')
    run_cell('a, b = x + 2, y + 3')
    run_cell('y = 43')
    run_cell('logging.info(a)')
    assert_not_detected('`a` does not depend on `y`')
    run_cell('logging.info(b)')
    assert_detected('`b` depends on stale `y`')
    run_cell('b = y + 10')
    run_cell('x = 99')
    run_cell('logging.info(b)')
    assert_not_detected('`b` does not depend on `x`')
    run_cell('logging.info(a)')
    assert_detected('`a` depends on stale `x`')


def test_attr_dep_with_top_level_overwrite():
    run_cell("""
class Foo:
    def __init__(self):
        self.y = 99
foo = Foo()
x = 42
foo.y = x + 7
""")
    run_cell('x = 43')
    run_cell('logging.info(foo)')
    assert_not_detected('only `foo.y` depends on x')
    run_cell('foo.y = 70')
    assert_not_detected('we just fixed stale dep of `foo.y` by changing its deps')
    run_cell('x = foo.y + 7')
    run_cell('foo = 81')
    run_cell('logging.info(x)')
    assert_detected('`x` has stale dep on `foo` (transitively through `foo.y`)')


def test_pandas():
    run_cell('import numpy as np')
    run_cell('import pandas as pd')
    run_cell('arr = 1 + np.arange(10)')
    run_cell('df = pd.DataFrame({"col": arr})')
    run_cell('df2 = df.dropna()')
    run_cell('df.dropna()')
    run_cell('logging.info(df2)')
    assert_not_detected('`df.dropna()` did not mutate `df`')
    run_cell('df.dropna(inplace=True)')
    run_cell('logging.info(df2)')
    assert_detected('`df.dropna(inplace=True)` mutated `df`')


@skipif_known_failing
def test_cell_magic():
    # TODO: write this
    pass



# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING

from .ipython_utils import cell_counter

if TYPE_CHECKING:
    from typing import Optional, Set


class DataCell(object):
    def __init__(
            self,
            name: str,
            obj_id: int,
            parents: 'Optional[Set[DataCell]]' = None,
    ):
        self.name = str(name)
        self.obj_id = obj_id
        if parents is None:
            parents = set()
        self.parents: Set[DataCell] = parents
        self.children: Set[DataCell] = set()
        self.defined_cell_num = cell_counter()

        # The notebook cell number this is required to have to not be considered stale
        self.required_cell_num = self.defined_cell_num

        # Set of ancestors defined more recently
        self.fresher_ancestors: Set[DataCell] = set()

        #Will never be stale if no_warning is True
        self.no_warning = False

    def __repr__(self):
        return f'<{self.__class__.__name__} for variable {self.name}>'

    def __str__(self):
        return self.name

    def update_deps(self, new_deps: 'Set[DataCell]', add=False, propagate_to_children=True):
        if not add:
            for parent in self.parents - new_deps:
                parent.children.discard(self)
            self.parents = set()

        for new_parent in new_deps - self.parents:
            new_parent.children.add(self)
            self.parents.add(new_parent)

        self.defined_cell_num = cell_counter()
        if propagate_to_children:
            for child in self.children:
                child._propagate_update(self)

    def mark_mutated(self, propagate_to_children=True):
        self.update_deps(set(), add=True, propagate_to_children=propagate_to_children)

    def _propagate_update(self, updated_dep: 'DataCell', seen=None):
        if seen is None:
            seen = set()
        if self in seen:
            return
        seen.add(self)
        self.required_cell_num = updated_dep.defined_cell_num
        self.fresher_ancestors.add(updated_dep)
        for child in self.children:
            child._propagate_update(updated_dep, seen=seen)

    def is_stale(self):
        if self.no_warning:
            return False
        return self.defined_cell_num < self.required_cell_num


class FunctionDataCell(DataCell):
    def __init__(self, scope, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scope = scope


class ClassDataCell(DataCell):
    def __init__(self, scope, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scope = scope



# -*- coding: utf-8 -*-
import inspect
from typing import TYPE_CHECKING

from IPython import get_ipython
try:
    import pandas
except ImportError:
    pandas = None

from .analysis import AttributeSymbolChain, CallPoint
from .data_cell import ClassDataCell, DataCell, FunctionDataCell

if TYPE_CHECKING:
    from typing import Dict, Optional, Set, Tuple, Union


class Scope(object):
    GLOBAL_SCOPE_NAME = '<module>'

    def __init__(
            self, scope_name: str = GLOBAL_SCOPE_NAME,
            parent_scope: 'Optional[Scope]' = None,
            is_namespace_scope=False,
    ):
        self.scope_name = scope_name
        self.parent_scope = parent_scope  # None iff this is the global scope
        self.cloned_from = None
        self.is_namespace_scope = is_namespace_scope
        self._data_cell_by_name: Dict[str, DataCell] = {}

    def __hash__(self):
        return hash(self.full_path)

    def __str__(self):
        return str(self.full_path)

    def clone(self):
        cloned = Scope()
        cloned.__dict__ = dict(self.__dict__)
        cloned.cloned_from = self
        cloned._data_cell_by_name = {}
        return cloned

    @property
    def non_namespace_parent_scope(self):
        # a scope nested inside of a namespace scope does not have access
        # to unqualified members of the namespace scope
        if self.is_global:
            return None
        if self.parent_scope.is_namespace_scope:
            return self.parent_scope.non_namespace_parent_scope
        return self.parent_scope

    def make_child_scope(self, scope_name, is_namespace_scope=False):
        return self.__class__(scope_name, parent_scope=self, is_namespace_scope=is_namespace_scope)

    def put(self, name: str, val: DataCell):
        self._data_cell_by_name[name] = val

    def lookup_data_cell_by_name_this_indentation(self, name):
        ret = self._data_cell_by_name.get(name, None)
        if ret is None and self.cloned_from is not None:
            ret = self.cloned_from.lookup_data_cell_by_name_this_indentation(name)
        return ret

    def all_data_cells_this_indentation(self):
        if self.cloned_from is None:
            ret = {}
        else:
            ret = self.cloned_from.all_data_cells_this_indentation()
        ret.update(self._data_cell_by_name)
        return ret

    def lookup_data_cell_by_name(self, name):
        ret = self.lookup_data_cell_by_name_this_indentation(name)
        if ret is None and self.non_namespace_parent_scope is not None:
            ret = self.non_namespace_parent_scope.lookup_data_cell_by_name(name)
        return ret

    def gen_data_cells_for_attr_symbol_chain(self, chain: AttributeSymbolChain, namespaces: 'Dict[int, Scope]'):
        cur_scope = self
        name_to_obj = get_ipython().ns_table['user_global']
        for name in chain.symbols:
            if isinstance(name, CallPoint):
                break
            dc = cur_scope.lookup_data_cell_by_name_this_indentation(name)
            if dc is not None:
                yield dc
            if name_to_obj is None:
                break
            obj = name_to_obj.get(name, None)
            if obj is None:
                break
            cur_scope = namespaces.get(id(obj), None)
            if cur_scope is None:
                break

            if (pandas is not None) and isinstance(obj, pandas.DataFrame):
                # FIXME: hack to get it working w/ pandas, which doesn't play nicely w/ inspect.getmembers
                name_to_obj = obj.__dict__
                name_to_obj.update(obj.to_dict())
            else:
                try:
                    name_to_obj = dict(inspect.getmembers(obj))
                except:  # noqa
                    name_to_obj = None

    def _upsert_and_mark_children_if_different_data_cell_type(
            self, dc: 'Union[ClassDataCell, FunctionDataCell]', name: str, deps: 'Set[DataCell]'
    ) -> 'Tuple[DataCell, DataCell, Optional[int]]':
        old_id = None
        old_dc = None
        if self.is_globally_accessible:
            old_dc = self.lookup_data_cell_by_name_this_indentation(name)
            if old_dc is not None:
                old_id = old_dc.obj_id
                # don't mark children as having stale dep unless old dep was of same type
                old_dc.update_deps(set(), add=False, propagate_to_children=isinstance(old_dc, type(dc)))
        dc.update_deps(deps, add=False)
        self.put(name, dc)
        return dc, old_dc, old_id

    def _upsert_function_data_cell_for_name(self, name: str, obj_id: int, deps: 'Set[DataCell]'):
        dc = FunctionDataCell(self.make_child_scope(name), name, obj_id)
        return self._upsert_and_mark_children_if_different_data_cell_type(dc, name, deps)

    def _upsert_class_data_cell_for_name(self, name: str, obj_id: int, deps: 'Set[DataCell]', class_scope: 'Scope'):
        dc = ClassDataCell(class_scope, name, obj_id)
        return self._upsert_and_mark_children_if_different_data_cell_type(dc, name, deps)

    def upsert_data_cell_for_name(
            self,
            name: str,
            obj_id: int,
            deps: 'Set[DataCell]',
            add=False,
            is_function_def=False,
            class_scope: 'Optional[Scope]' = None,
    ) -> 'Tuple[DataCell, DataCell, Optional[int]]':
        assert not (class_scope is not None and is_function_def)
        if is_function_def:
            assert not add
            return self._upsert_function_data_cell_for_name(name, obj_id, deps)
        if class_scope is not None:
            assert not add
            return self._upsert_class_data_cell_for_name(name, obj_id, deps, class_scope)
        old_id = None
        old_dc = None
        if self.is_globally_accessible:
            old_dc = self.lookup_data_cell_by_name_this_indentation(name)
            if old_dc is not None:
                old_id = old_dc.obj_id
                # TODO: garbage collect old names
                # TODO: handle case where new dc is of different type
                if name in self._data_cell_by_name:
                    old_dc.update_deps(deps, add=add)
                    old_dc.obj_id = obj_id
                    return old_dc, old_dc, old_id
                else:
                    # in this case, we are copying from a class and should add the dc from which we are copying
                    # as an additional dependency
                    deps.add(old_dc)
        dc = DataCell(name, obj_id, deps)
        self.put(name, dc)
        for dep in deps:
            dep.children.add(dc)
        return dc, old_dc, old_id

    @property
    def is_global(self):
        return self.parent_scope is None

    @property
    def is_globally_accessible(self):
        return self.is_global or (self.is_namespace_scope and self.parent_scope.is_globally_accessible)

    @property
    def global_scope(self):
        if self.is_global:
            return self
        return self.parent_scope.global_scope

    @property
    def full_path(self) -> 'Tuple[str, ...]':
        path = (self.scope_name,)
        if self.is_global:
            return path
        else:
            return self.parent_scope.full_path + path



# -*- coding: utf-8 -*-
import ast
from contextlib import contextmanager
import logging
from typing import TYPE_CHECKING

from ..analysis import get_statement_lval_and_rval_symbols
from ..data_cell import FunctionDataCell

if TYPE_CHECKING:
    from types import FrameType
    from typing import List, Optional, Set
    from ..data_cell import DataCell
    from ..safety import DependencySafety
    from ..scope import Scope

logger = logging.getLogger(__name__)


class TraceStatement(object):
    def __init__(self, safety: 'DependencySafety', frame: 'FrameType', stmt_node: 'ast.stmt', scope: 'Scope'):
        self.safety = safety
        self.frame = frame
        self.stmt_node = stmt_node
        self.scope = scope
        self.class_scope: Optional[Scope] = None
        self.call_point_dependencies: List[Set[DataCell]] = []
        self.marked_finished = False

    @contextmanager
    def replace_active_scope(self, new_active_scope):
        old_scope = self.scope
        self.scope = new_active_scope
        yield
        self.scope = old_scope

    def compute_rval_dependencies(self, rval_symbols=None):
        if rval_symbols is None:
            _, rval_symbols, _ = get_statement_lval_and_rval_symbols(self.stmt_node)
        rval_data_cells = set()
        for name in rval_symbols:
            maybe_rval_dc = self.scope.lookup_data_cell_by_name(name)
            if maybe_rval_dc is not None:
                rval_data_cells.add(maybe_rval_dc)
        return rval_data_cells.union(*self.call_point_dependencies) | self.safety.attr_trace_manager.loaded_data_cells

    def get_post_call_scope(self, old_scope: 'Scope'):
        if isinstance(self.stmt_node, ast.ClassDef):
            # classes need a new scope before the ClassDef has finished executing,
            # so we make it immediately
            return self.scope.make_child_scope(self.stmt_node.name, is_namespace_scope=True)

        if not isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            # TODO: probably the right thing is to check is whether a lambda appears somewhere inside the ast node
            # if not isinstance(self.ast_node, ast.Lambda):
            #     raise TypeError('unexpected type for ast node %s' % self.ast_node)
            return old_scope
        func_name = self.stmt_node.name
        func_cell = self.scope.lookup_data_cell_by_name(func_name)
        if func_cell is None:
            # TODO: brittle; assumes any user-defined and traceable function will always be present; is this safe?
            return old_scope
        if not isinstance(func_cell, FunctionDataCell):
            raise TypeError('got non-function data cell %s for name %s' % (func_cell, func_name))
        return func_cell.scope

    def _get_obj_id_for_name(self, name):
        try:
            return id(self.frame.f_locals[name])
        except KeyError:
            logger.error('unable to find object for name %s', name)
            return id(None)

    def _handle_aliases(
            self,
            old_id: 'Optional[int]', old_dc: 'Optional[DataCell]',
            obj_id: 'Optional[int]', dc: 'Optional[DataCell]'
    ):
        if old_id is not None and old_dc is not None:
            self.safety.aliases[old_id].discard(old_dc)
        if obj_id is not None and dc is not None:
            self.safety.aliases[obj_id].add(dc)
        if old_id == obj_id:
            for alias_dc in self.safety.aliases[obj_id]:
                alias_dc.update_deps(set(), add=True)

    def _make_lval_data_cells(self):
        lval_symbols, rval_symbols, should_add = get_statement_lval_and_rval_symbols(self.stmt_node)
        rval_deps = self.compute_rval_dependencies(rval_symbols=rval_symbols - lval_symbols)
        is_function_def = isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef))
        is_class_def = isinstance(self.stmt_node, ast.ClassDef)
        if is_function_def or is_class_def:
            assert len(lval_symbols) == 1
            assert not lval_symbols.issubset(rval_symbols)
        for name in lval_symbols:
            should_add_for_name = should_add or name in rval_symbols
            if is_class_def:
                assert self.class_scope is not None
                class_ref = self.frame.f_locals[self.stmt_node.name]
                self.safety.namespaces[id(class_ref)] = self.class_scope
            # if is_function_def:
            #     print('create function', name, 'in scope', self.scope)
            obj_id = self._get_obj_id_for_name(name)
            dc, old_dc, old_id = self.scope.upsert_data_cell_for_name(
                name, obj_id, rval_deps,
                add=should_add_for_name, is_function_def=is_function_def, class_scope=self.class_scope
            )
            self._handle_aliases(old_id, old_dc, obj_id, dc)
        if len(self.safety.attr_trace_manager.saved_store_data) > 0:
            assert isinstance(self.stmt_node, ast.Assign)
        if len(self.safety.attr_trace_manager.saved_aug_store_data) > 0:
            assert isinstance(self.stmt_node, ast.AugAssign)
        for scope, obj, attr in self.safety.attr_trace_manager.saved_store_data:
            obj_id = id(getattr(obj, attr, None))
            dc, old_dc, old_id = scope.upsert_data_cell_for_name(
                attr, obj_id, rval_deps, add=False, is_function_def=False, class_scope=None
            )
            self._handle_aliases(old_id, old_dc, obj_id, dc)
        for scope, obj, attr in self.safety.attr_trace_manager.saved_aug_store_data:
            obj_id = id(getattr(obj, attr, None))
            dc, old_dc, old_id = scope.upsert_data_cell_for_name(
                attr, obj_id, rval_deps, add=True, is_function_def=False, class_scope=None
            )
            self._handle_aliases(old_id, old_dc, obj_id, dc)

    def handle_dependencies(self):
        if not self.safety.dependency_tracking_enabled:
            return
        for mutated_obj_id, mutation_args in self.safety.attr_trace_manager.mutations:
            mutation_arg_dcs = set(self.scope.lookup_data_cell_by_name(arg) for arg in mutation_args) - {None}
            for mutated_dc in self.safety.aliases[mutated_obj_id]:
                mutated_dc.update_deps(mutation_arg_dcs, add=True)
        if self.has_lval:
            self._make_lval_data_cells()
        else:
            assert len(self.safety.attr_trace_manager.saved_store_data) == 0
            assert len(self.safety.attr_trace_manager.saved_aug_store_data) == 0

    def finished_execution_hook(self):
        if self.marked_finished:
            return
        # print('finishing stmt', self.stmt_node)
        self.marked_finished = True
        self.handle_dependencies()

    @property
    def has_lval(self):
        # TODO: expand to method calls, etc.
        return isinstance(self.stmt_node, (
            ast.Assign, ast.AugAssign, ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef, ast.For
        ))



# -*- coding: utf-8 -*-


class CommonEqualityMixin(object):
    def __eq__(self, other):
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not (self == other)
